"""STH specific test code

Use this test code in addition to the one for the sensor node:

    icotest run -k 'sensor_node or sth'

"""

# -- Imports ------------------------------------------------------------------

from logging import getLogger

from icotronic.can import STH, StreamingConfiguration
from icotronic.measurement.constants import ADC_MAX_VALUE
from icotronic.measurement import convert_raw_to_g, ratio_noise_max
from pytest import mark

from icotest.config import settings
from icotest.test.support.node import check_write_read_eeprom_close
from icotest.test.support.sensor_node import read_streaming_data
from icotest.test.support.sth import read_self_test_voltages

# -- Functions ----------------------------------------------------------------


@mark.anyio
async def test_acceleration_sensor_self_test(sth: STH):
    """Use the self test of a acceleration sensor to check for problems"""

    voltage_diff_abs, voltage_diff_before_after = (
        await read_self_test_voltages(sth)
    )

    sensor = settings.acceleration_sensor()

    voltage_diff_expected = sensor.self_test.voltage.difference
    voltage_diff_tolerance = sensor.self_test.voltage.tolerance

    voltage_diff_minimum = voltage_diff_expected - voltage_diff_tolerance
    voltage_diff_maximum = voltage_diff_expected + voltage_diff_tolerance

    assert voltage_diff_before_after <= voltage_diff_tolerance, (
        "Measured voltage difference between voltage before and after "
        f"test {voltage_diff_before_after:.2f} mV is larger than "
        f"tolerance of {voltage_diff_tolerance:.2f} mV"
    )
    possible_failure_reason = (
        "\n\nPossible Reason:\n\n• Acceleration sensor config value "
        f"“{settings.sth.acceleration_sensor.sensor}” is incorrect"
    )

    assert voltage_diff_minimum <= voltage_diff_abs, (
        f"Measured voltage difference of {voltage_diff_abs:.2f} mV is "
        "lower than expected minimum voltage difference of "
        f"{voltage_diff_minimum:.2f} mV{possible_failure_reason}"
    )
    assert voltage_diff_abs <= voltage_diff_maximum, (
        f"Measured voltage difference of {voltage_diff_abs:.2f} mV is "
        "greater than expected minimum voltage difference of "
        f"{voltage_diff_maximum:.2f} mV{possible_failure_reason}"
    )


@mark.anyio
async def test_acceleration_single_value(sth: STH):
    """Test stationary acceleration value"""

    stream_data = await sth.get_streaming_data_single()
    sensor = settings.acceleration_sensor()
    acceleration = convert_raw_to_g(
        stream_data.values[0], sensor.acceleration.maximum
    )

    logger = getLogger(__file__)
    logger.info("Measured acceleration value: %.2f g", acceleration)

    # We expect a stationary acceleration between -g₀ and g₀ (g₀ = 9.807 m/s²)
    expected_acceleration = 0
    tolerance_acceleration = sensor.acceleration.tolerance
    expected_minimum_acceleration = (
        expected_acceleration - tolerance_acceleration
    )
    expected_maximum_acceleration = (
        expected_acceleration + tolerance_acceleration
    )

    assert expected_minimum_acceleration <= acceleration, (
        f"Measured acceleration {acceleration:.3f} g is lower "
        "than expected minimum acceleration "
        f"{expected_minimum_acceleration:.3f} g"
    )
    assert acceleration <= expected_maximum_acceleration, (
        f"Measured acceleration {acceleration:.3f} g is greater "
        "than expected maximum acceleration "
        f"{expected_maximum_acceleration:.3f} g"
    )


@mark.anyio
async def test_acceleration_noise(sth: STH):
    """Test ratio of noise to maximal possible measurement value"""

    acceleration = await read_streaming_data(
        sth, StreamingConfiguration(first=True), length=10000
    )

    ratio_noise_maximum = ratio_noise_max(acceleration)
    sensor = settings.acceleration_sensor()
    maximum_ratio_allowed = sensor.acceleration.ratio_noise_to_max_value

    assert ratio_noise_maximum <= maximum_ratio_allowed, (
        "The ratio noise to possible maximum measured value of "
        f"{ratio_noise_maximum} dB is higher than the maximum allowed level "
        f"of {maximum_ratio_allowed} dB"
    )


@mark.anyio
async def test_eeprom(sth: STH):
    """Test if reading and writing STH EEPROM data works"""

    sensor = settings.acceleration_sensor()
    acceleration_max = sensor.acceleration.maximum

    acceleration_slope = acceleration_max / ADC_MAX_VALUE
    acceleration_offset = -(acceleration_max / 2)

    for axis in ("x", "y", "z"):
        await check_write_read_eeprom_close(
            sth, f"{axis} axis acceleration slope", acceleration_slope
        )
        await check_write_read_eeprom_close(
            sth, f"{axis} axis acceleration offset", acceleration_offset
        )



"""Support code for handling device software"""

# -- Imports ------------------------------------------------------------------

from pathlib import Path

from icotest.cli.commander import Commander

# -- Functions ----------------------------------------------------------------


def upload_flash(chip: str, flash_location: str | Path):
    """Upload bootloader and application into node

    Args:

        chip:
            The name of the chip that should be flashed

        flash_location:
            The location of the flash image

    """

    image_filepath = Path(flash_location).expanduser().resolve()
    if not image_filepath.exists():
        raise FileNotFoundError(
            "Firmware file {image_filepath} does not exist"
        )
    if not image_filepath.is_file():
        raise FileNotFoundError(
            f"Firmware file {image_filepath} is not a file"
        )

    Commander().upload_flash(chip=chip, filepath=image_filepath)



"""Common test support code"""

# -- Imports ------------------------------------------------------------------

from dynaconf.utils.boxing import DynaBox

# -- Functions ----------------------------------------------------------------


async def check_power_usage(power_usage: float, settings: DynaBox):
    """Test if the average power usage matches a certain value

    Args:

        power_usage:

            The measured power usage in mW

        settings:

            The settings object that contains the expected power usage values

    """

    average_power = settings.average
    tolerance = settings.tolerance

    minimum_power = average_power - tolerance
    maximum_power = average_power + tolerance
    assert minimum_power <= power_usage, (
        f"Power usage of {power_usage} mW smaller than expected minimum of "
        f"{minimum_power} mW"
    )
    assert power_usage <= maximum_power, (
        f"Power usage of {power_usage} mW larger than expected maximum of "
        f"{maximum_power} mW"
    )



"""Test sensor node hardware (SHA, STH, SMH…)"""

# -- Imports ------------------------------------------------------------------

from asyncio import Event, TaskGroup, to_thread
from logging import getLogger

from icotronic.can import SensorNode, StreamingConfiguration, STU
from pytest import mark

from icotest.cli.commander import Commander
from icotest.config import settings
from icotest.test.support.common import check_power_usage
from icotest.test.support.node import (
    check_connection,
    check_firmware_upload,
    check_eeprom_product_data,
    check_eeprom_statistics,
    check_eeprom_status,
)
from icotest.test.support.sensor_node import (
    check_eeprom_name,
    check_eeprom_bluetooth_times,
)

# -- Functions ----------------------------------------------------------------


@mark.anyio
async def test_firmware_upload():
    """Upload firmware"""

    await check_firmware_upload(settings.sensor_node)


@mark.anyio
async def test_connection(sensor_node: SensorNode):
    """Test if connection to sensor node is possible"""

    await check_connection(sensor_node)


@mark.anyio
async def test_supply_voltage(sensor_node: SensorNode):
    """Test if battery voltage is within expected bounds"""

    supply_voltage = await sensor_node.get_supply_voltage()
    expected_voltage = settings.sensor_node.supply.voltage.average
    tolerance_voltage = settings.sensor_node.supply.voltage.tolerance

    expected_minimum_voltage = expected_voltage - tolerance_voltage
    expected_maximum_voltage = expected_voltage + tolerance_voltage

    assert supply_voltage >= expected_minimum_voltage, (
        (
            f"Supply voltage of {supply_voltage:.3f} V is lower "
            "than expected minimum voltage of "
            f"{expected_minimum_voltage:.3f} V"
        ),
    )
    assert supply_voltage <= expected_maximum_voltage, (
        (
            f"Supply voltage of {supply_voltage:.3f} V is "
            "greater than expected maximum voltage of "
            f"{expected_minimum_voltage:.3f} V"
        ),
    )


@mark.anyio
async def test_power_usage_disconnected(
    stu: STU,  # pylint: disable=unused-argument
) -> None:
    """Check power usage in disconnected state"""

    commander = Commander()
    commander.enable_debug_mode()
    power_usage_mw = commander.read_power_usage()
    getLogger(__name__).info("Disconnected power usage: %s mW", power_usage_mw)

    await check_power_usage(
        power_usage_mw, settings.sensor_node.power.disconnected
    )


@mark.anyio
async def test_power_usage_connected(
    sensor_node: SensorNode,  # pylint: disable=unused-argument
) -> None:
    """Check power usage in connected state"""

    commander = Commander()
    commander.enable_debug_mode()
    power_usage_mw = commander.read_power_usage()
    getLogger(__name__).info("Connected power usage: %s mW", power_usage_mw)

    await check_power_usage(
        power_usage_mw, settings.sensor_node.power.connected
    )


@mark.anyio
async def test_power_usage_streaming(sensor_node: SensorNode):
    """Test power usage of sensor node while streaming"""

    async def stream_data(started_streaming: Event) -> None:
        async with sensor_node.open_data_stream(
            StreamingConfiguration(first=True)
        ) as stream:
            async for _ in stream:
                if not started_streaming.is_set():
                    started_streaming.set()

    def read_power_usage() -> float:
        return Commander().read_power_usage()

    started_streaming = Event()

    async with TaskGroup() as task_group:
        stream_data_task = task_group.create_task(
            stream_data(started_streaming)
        )
        await started_streaming.wait()
        read_power_task = task_group.create_task(to_thread(read_power_usage))
        power_usage_mw = await read_power_task
        getLogger(__name__).info(
            "Streaming power usage: %s mW", power_usage_mw
        )
        stream_data_task.cancel()

    await check_power_usage(
        power_usage_mw, settings.sensor_node.power.streaming
    )


@mark.anyio
async def test_eeprom(sensor_node: SensorNode):
    "Test if reading and writing of EEPROM values works"

    await check_eeprom_name(sensor_node, settings.sensor_node)
    await check_eeprom_product_data(sensor_node, settings.sensor_node)
    await check_eeprom_statistics(sensor_node, settings.sensor_node)
    await check_eeprom_status(sensor_node)
    await check_eeprom_bluetooth_times(sensor_node, settings.sensor_node)



"""Configuration for pytest"""

# -- Imports ------------------------------------------------------------------

from pytest import fixture

from icotronic.can import Connection, SensorNode, STH, STU
from netaddr import EUI

from icotest.config import settings

# pylint: disable=redefined-outer-name

# -- Fixtures -----------------------------------------------------------------


@fixture(scope="session")
def anyio_backend():
    """Set default async backend"""

    return "asyncio"


@fixture(scope="session")
def sensor_node_name() -> str:
    """Returns the name of the sensor node used for the test"""

    return settings.sensor_node.name


@fixture(scope="session")
async def sensor_node_mac_address(sensor_node_name: str) -> EUI:
    """Return the MAC address of the sensor node used for the test"""

    async with Connection() as stu:
        async with stu.connect_sensor_node(sensor_node_name) as sensor_node:
            return await sensor_node.get_mac_address()


@fixture
async def stu() -> STU:
    """Connect to and disconnect from STU"""

    async with Connection() as stu:
        yield stu


@fixture
async def sensor_node(stu, sensor_node_name) -> SensorNode:
    """Connect to and disconnect from sensor node"""

    async with stu.connect_sensor_node(sensor_node_name) as sensor_node:
        yield sensor_node


@fixture
async def sth(stu, sensor_node_name) -> STH:
    """Connect to and disconnect from an STH"""

    async with stu.connect_sensor_node(sensor_node_name, STH) as sth:
        yield sth



"""Command line tool for hardware tests"""

# -- Imports ------------------------------------------------------------------

from argparse import ArgumentParser
from logging import basicConfig, getLogger
from subprocess import run, CalledProcessError
from sys import exit as sys_exit

from icotest.config import ConfigurationUtility

# -- Functions ----------------------------------------------------------------


def create_icotest_parser() -> ArgumentParser:
    """Create command line parser for ICOtest

    Returns:

        A parser for the CLI arguments of icotest

    """

    parser = ArgumentParser(description="ICOtest CLI tool")

    parser.add_argument(
        "--log",
        choices=("debug", "info", "warning", "error", "critical"),
        default="warning",
        required=False,
        help="minimum log level",
    )

    subparsers = parser.add_subparsers(
        required=True, title="Subcommands", dest="subcommand"
    )

    # ==========
    # = Config =
    # ==========

    subparsers.add_parser(
        "config", help="Open configuration file in default application"
    )

    # =======
    # = Run =
    # =======

    subparsers.add_parser("run", help="Run tests")

    return parser


def run_pytest(log_level: str, pytest_args: list[str]) -> None:
    """Run pytest for the package using the given arguments

    Args:

        log_level:

            Log level for invocation of pytest

        pytest_args:

            Additional arguments for pytest call

    """

    command = [
        "pytest",
        "--log-cli-level",
        log_level,
        "--pyargs",
        "icotest.test",
    ] + pytest_args
    print(f"\nTest Command:\n\n  {' '.join(command)}\n")
    try:
        run(command, check=True)
    except CalledProcessError as error:
        sys_exit(error.returncode)


# -- Main ---------------------------------------------------------------------


def main() -> None:
    """ICOtest command line tool"""

    parser = create_icotest_parser()
    # Parse known args to get subcommand
    arguments, additional_args = parser.parse_known_args()
    if vars(arguments).get("subcommand", "undefined") != "run":
        arguments = parser.parse_args()

    log_level = arguments.log.upper()
    basicConfig(
        level=log_level,
        style="{",
        format="{asctime} {levelname:7} {message}",
    )

    logger = getLogger(__name__)
    logger.info("CLI arguments: %s", arguments)
    logger.info("Additional unrecognized arguments: %s", additional_args)

    subcommand = arguments.subcommand

    match subcommand:
        case "config":
            ConfigurationUtility.open_user_config()
        case "run":
            run_pytest(log_level, additional_args)


if __name__ == "__main__":
    main()



"""Test STU"""

# -- Imports ------------------------------------------------------------------

from icotronic.can import STU
from pytest import mark

from icotest.config import settings
from icotest.test.support.node import (
    check_firmware_upload,
    check_connection,
    check_eeprom_product_data,
    check_eeprom_statistics,
    check_eeprom_status,
)

# -- Functions ----------------------------------------------------------------


@mark.anyio
async def test_firmware_upload():
    """Upload firmware"""

    await check_firmware_upload(settings.stu)


@mark.anyio
async def test_connection(stu: STU):
    """Test if connection to STU is possible"""

    await check_connection(stu)


@mark.anyio
async def test_eeprom(stu: STU):
    "Test if reading and writing of EEPROM values works"

    await check_eeprom_product_data(stu, settings.stu)
    await check_eeprom_statistics(stu, settings.stu)
    await check_eeprom_status(stu)



"""Shared code for all sensor nodes of the ICOtronic system"""

# -- Imports ------------------------------------------------------------------

from dynaconf.utils.boxing import DynaBox
from icotronic.can import SensorNode, StreamingConfiguration

from icotest.test.support.node import check_write_read_eeprom

# -- Functions ----------------------------------------------------------------


async def check_eeprom_name(node: SensorNode, settings: DynaBox):
    """Test if reading and writing the name into the EEPROM works

    Args:

        node:

            The sensor node that should be checked

        settings:

            The settings object that contains the sensor node setting

    """

    await check_write_read_eeprom(node, "name", settings.name)


async def check_eeprom_bluetooth_times(node: SensorNode, settings: DynaBox):
    """Test if reading and writing the Bluetooth times works

    Args:

        node:

            The sensor node that should be checked

        settings:

            The settings object that contains the sensor node setting

    """

    bluetooth = settings.bluetooth

    await check_write_read_eeprom(
        node, "advertisement time 1", bluetooth.advertisement_time_1
    )
    await check_write_read_eeprom(node, "sleep time 1", bluetooth.sleep_time_1)
    await check_write_read_eeprom(
        node, "advertisement time 2", bluetooth.advertisement_time_2
    )
    await check_write_read_eeprom(node, "sleep time 2", bluetooth.sleep_time_2)


async def read_streaming_data(
    node: SensorNode, config: StreamingConfiguration, length: int
) -> list[float]:
    """Collect a certain amount of streaming data

    Args:

        node:

            The sensor node where streaming data should be collected

        config:

            The streaming configuration that should be used to collect data

        length:

            The amount of streaming values that should be collected

    Returns:

        A list of containing raw 16 bit streaming values

    """

    async with node.open_data_stream(config) as stream:
        stream_data = []
        async for data, _ in stream:
            stream_data.extend(data.values)
            if len(stream_data) >= length:
                break

    # Due to the chosen streaming format the code above might have
    # collected one or two additional values. We remove these values
    # here.
    assert len(stream_data) >= length
    additional_values = len(stream_data) - length
    return stream_data[:-additional_values]



"""Shared code for all nodes of the ICOtronic system (STU & sensor nodes)"""

# -- Imports ------------------------------------------------------------------

from asyncio import sleep
from logging import getLogger
from math import isclose
from operator import __eq__
from typing import Callable, TypeVar

from dynaconf.utils.boxing import DynaBox
from icotronic.can import SensorNode, STH, STU
from icotronic.can.node.eeprom.status import EEPROMStatus
from icotronic.can.status import State
from semantic_version import Version

from icotest.firmware import upload_flash

# -- Types --------------------------------------------------------------------

EEPROMValue = TypeVar("EEPROMValue", Version, str)
"""Type of an object that can be written into EEPROM"""

# -- Functions ----------------------------------------------------------------


async def check_firmware_upload(node_settings: DynaBox):
    """Upload firmware"""

    logger = getLogger(__name__)
    firmware_location = node_settings.firmware.location
    logger.info("Firmware Location: %s", firmware_location)

    chip = node_settings.firmware.chip
    upload_flash(chip, firmware_location)


async def check_connection(node: SensorNode | STU) -> None:
    """Check connection to node

    Args:

        node:

            The node for that should be checked

    """

    await sleep(1)  # Wait for startup of sensor node

    # Just send a request for the state and check, if the result
    # matches our expectations.
    state = await node.get_state()

    expected_state = State(
        mode="Get", location="Application", state="Operating"
    )

    assert state == expected_state, (
        (
            f"Expected state “{expected_state}” does not match "
            f"received state “{state}”"
        ),
    )


async def check_write_read_eeprom_function(
    node: SensorNode | STH | STU,
    name: str,
    written: EEPROMValue,
    comparator: Callable[[EEPROMValue, EEPROMValue], bool],
    description: str,
):
    """Check a written and read EEPROM value for relation to each other

    Args:

        node:
                The node that should be checked

        name:

                The name of the EEPROM value

        written:

                The value that should be written and then read afterwards

        comparator:

                The function that will be applied to check if the values are
                related or not

        description:

                A text that describes the **inverse** of the comparator
                function; e.g. for something like the function ``__eq__`` this
                could be something like ``"is not equal to"``.

    """

    function_name = name.lower().replace(" ", "_")
    write_coroutine = getattr(node.eeprom, f"write_{function_name}")
    read_coroutine = getattr(node.eeprom, f"read_{function_name}")
    await write_coroutine(written)
    read = await read_coroutine()
    logger = getLogger(__name__)
    logger.debug("Type of written value: %s", type(written))
    logger.debug("Type of read value: %s", type(read))
    assert comparator(
        written, read
    ), f"Written {name} “{written}” {description} read {name} “{read}”"


async def check_write_read_eeprom(
    node: SensorNode | STH | STU, name: str, written: EEPROMValue
) -> None:
    """Check that a written and read EEPROM value match

    Args:

        node:
                The node that should be checked

        name:

                The name of the EEPROM value

        written:

                The value that should be written and then read afterwards

    """

    return await check_write_read_eeprom_function(
        node, name, written, __eq__, "does not match"
    )


async def check_write_read_eeprom_close(
    node: SensorNode | STH | STU, name: str, written: float
) -> None:
    """Check that a written and read EEPROM value are approximately the same

    Args:

        node:
                The node that should be checked

        name:

                The name of the EEPROM value

        written:

                The value that should be written and then read afterwards

    """

    return await check_write_read_eeprom_function(
        node, name, written, isclose, "is not close to"
    )


async def check_eeprom_product_data(node: SensorNode | STU, settings: DynaBox):
    """Test if reading and writing EEPROM product data works

    Args:

        node:

            The node that should be checked

        settings:

            The settings object that contains the node setting

    """

    await check_write_read_eeprom(node, "GTIN", settings.gtin)
    await check_write_read_eeprom(
        node, "hardware version", Version.coerce(settings.hardware_version)
    )
    # I am not sure, if the firmware already inits the EEPROM with the firmware
    # version. Writing back the same firmware version into the EEPROM should
    # not be a problem though.
    await check_write_read_eeprom(
        node, "firmware version", await node.get_firmware_version()
    )
    # Originally we assumed that this value would be set by the
    # firmware itself. However, according to tests with an empty EEPROM
    # this is not the case.
    await check_write_read_eeprom(
        node, "release name", settings.firmware.release_name
    )
    await check_write_read_eeprom(
        node, "serial number", settings.serial_number
    )
    await check_write_read_eeprom(node, "product name", settings.product_name)
    await check_write_read_eeprom(node, "OEM data", settings.oem_data)


async def check_eeprom_statistics(node: SensorNode | STU, settings: DynaBox):
    """Test if reading and writing EEPROM statistic data works

    Args:

        node:

            The node that should be checked

        settings:

            The settings object that contains the node setting

    """

    for attribute in (
        "power on cycles",
        "power off cycles",
        "operating time",
        "under voltage counter",
        "watchdog reset counter",
    ):
        await check_write_read_eeprom(node, attribute, 0)

    await check_write_read_eeprom(
        node, "production date", settings.production_date
    )
    await check_write_read_eeprom(node, "batch number", settings.batch_number)


async def check_eeprom_status(node: SensorNode | STU):
    """Test if reading and writing the EEPROM status byte works

    Args:

        node:

            The node that should be checked

    """

    await check_write_read_eeprom(node, "status", EEPROMStatus("Initialized"))



"""Support for changing configuration values for the tests"""

# -- Import -------------------------------------------------------------------

from datetime import date
from functools import partial
from importlib.resources import as_file, files
from numbers import Real
from pathlib import Path
from sys import exit as sys_exit, stderr


from dynaconf import (  # type: ignore[attr-defined]
    Dynaconf,
    ValidationError,
    Validator,
)

from dynaconf.vendor.ruamel.yaml.parser import ParserError
from dynaconf.vendor.ruamel.yaml.scanner import ScannerError

from startfile import startfile
from platformdirs import site_config_dir, user_config_dir

# -- Functions ----------------------------------------------------------------


def must_exist(*arguments, **keyword_arguments) -> Validator:
    """Return Validator which requires setting to exist

    Args:

        arguments:

            All positional arguments for the Validator

        keyword_arguments:

            All keyword arguments for the Validator

    """

    return Validator(*arguments, must_exist=True, **keyword_arguments)


def element_is_type(nodes, name: str, element_type: type):
    """Check that all elements of a list have a certain type

    Args:

        nodes:

            The parent node of the list that should be checked

        name:

            The name of the parent node

        element_type:

            The expected type of the nodes in the list

    Returns:

        ``True``, if every element has the expected type

    Raises:

        ``ValidationError``, if any element of the given list has the wrong
        type

    """

    if nodes is None:
        return True  # Let parent validator handle wrong type

    for node in nodes:
        if not isinstance(node, element_type):
            raise ValidationError(
                f"Element “{node}” of {name} has wrong type "
                f"“{type(node)}” instead of “{element_type}”"
            )
    return True


def element_is_int(nodes, name: str):
    """Check that all elements of a list are numbers

    Args:

        nodes:

            The parent node of the list that should be checked

        name:

            The name of the parent node

    Returns:

        ``True``, if every element has the type ``int``

    Raises:

        ``ValidationError``, if any element of the given list has a type other
        than ``int``

    """

    return element_is_type(nodes, name, element_type=int)


def element_is_string(nodes, name: str):
    """Check that all elements of a list are strings

    Args:

        nodes:

            The parent node of the list that should be checked

        name:

            The name of the parent node

    Returns:

        ``True``, if every element has the type ``str``

    Raises:

        ``ValidationError``, if any element of the given list has a type other
        than ``str``

    """

    return element_is_type(nodes, name, element_type=str)


def commands_validators() -> list[Validator]:
    """Return list of validators for config data below key `commands`"""

    return [
        must_exist(
            f"commands.path.{os}",
            is_type_of=list,
            condition=partial(element_is_string, name=f"commands.path.{os}"),
        )
        for os in ("linux", "mac", "windows")
    ]


def acceleration_sensor_validators(name: str):
    """Return the list of validators for a specific acceleration sensor

    Args:

        name:

            The name of the acceleration sensor

    """

    prefix = "sth.acceleration_sensor"
    return [
        must_exist(
            f"{prefix}.{name}.acceleration.maximum",
            f"{prefix}.{name}.acceleration.tolerance",
            f"{prefix}.{name}.reference_voltage",
            f"{prefix}.{name}.self_test.voltage.difference",
            f"{prefix}.{name}.self_test.voltage.tolerance",
            is_type_of=Real,
        ),
        must_exist(
            f"{prefix}.{name}.self_test.dimension",
            is_type_of=str,
            is_in=("x", "y", "z"),
        ),
    ]


def node_validators(node: str) -> list[Validator]:
    """Get validators for node (STU & sensor node) configuration

    Args:

        node:

            The namespace of the node configuration

    Returns:

        A list of validators for the given node

    """

    return [
        must_exist(f"{node}.gtin", f"{node}.batch_number", is_type_of=int),
        must_exist(
            f"{node}.hardware_version",
            f"{node}.firmware.chip",
            f"{node}.firmware.location",
            f"{node}.firmware.release_name",
            f"{node}.product_name",
            f"{node}.serial_number",
            is_type_of=str,
        ),
        must_exist(
            f"{node}.oem_data",
            is_type_of=list,
            condition=partial(element_is_int, name="{node}.oem_data"),
        ),
        must_exist(
            f"{node}.production_date",
            is_type_of=date,
        ),
    ]


def sensor_node_validators() -> list[Validator]:
    """Return list of validators for config data below key `sensor_node`"""

    return node_validators("sensor_node") + [
        must_exist(
            "sensor_node.name",
            is_type_of=str,
        ),
        must_exist(
            "sensor_node.supply.voltage.average",
            "sensor_node.supply.voltage.tolerance",
            "sensor_node.power.connected.average",
            "sensor_node.power.connected.tolerance",
            "sensor_node.power.disconnected.average",
            "sensor_node.power.disconnected.tolerance",
            "sensor_node.power.streaming.average",
            "sensor_node.power.streaming.tolerance",
            is_type_of=Real,
        ),
        must_exist(
            "sensor_node.bluetooth.advertisement_time_1",
            "sensor_node.bluetooth.advertisement_time_2",
            "sensor_node.bluetooth.sleep_time_1",
            "sensor_node.bluetooth.sleep_time_2",
            is_type_of=int,
        ),
    ]


def stu_validators() -> list[Validator]:
    """Return list of validators for config data below key `stu`"""

    return node_validators("stu")


def sth_validators() -> list[Validator]:
    """Return list of validators for config data below key `sth`"""

    validators = []
    for sensor in ("ADXL1001", "ADXL1002", "ADXL356"):
        validators.extend(acceleration_sensor_validators(sensor))

    return validators


def handle_incorrect_settings(error_message: str) -> None:
    """Handle incorrect configuration

    Args:

        error_message:

            A text that describes the configuration error

    """

    print(error_message, file=stderr)
    print(
        "\n"
        "• Most likely this problem is caused by an incorrect user "
        "configuration.\n"
        "• Please fix the problem and try again afterwards.\n\n"
        "Opening your user config file in your text editor now",
        file=stderr,
    )
    ConfigurationUtility.open_user_config()
    sys_exit(1)


# -- Classes ------------------------------------------------------------------


class ConfigurationUtility:
    """Access configuration data"""

    app_name = "ICOtest"
    app_author = "MyTooliT"
    config_filename = "config.yaml"
    site_config_filepath = (
        Path(site_config_dir(app_name, appauthor=app_author)) / config_filename
    )
    user_config_filepath = (
        Path(user_config_dir(app_name, appauthor=app_author)) / config_filename
    )

    @staticmethod
    def open_config_file(filepath: Path):
        """Open configuration file

        Args:

            filepath:
                Path to configuration file

        """

        # Create file, if it does not exist already
        if not filepath.exists():
            filepath.parent.mkdir(
                exist_ok=True,
                parents=True,
            )

            default_user_config = (
                files("icotest.config")
                .joinpath("user.yaml")
                .read_text(encoding="utf-8")
            )

            with filepath.open("w", encoding="utf8") as config_file:
                config_file.write(default_user_config)

        startfile(filepath)

    @classmethod
    def open_user_config(cls):
        """Open the current users configuration file"""

        try:
            cls.open_config_file(cls.user_config_filepath)
        except FileNotFoundError as error:
            print(
                f"Unable to open user configuration: {error}"
                "\nTo work around this problem please open "
                f"“{cls.user_config_filepath}” in your favorite text "
                "editor",
                file=stderr,
            )


class SettingsIncorrectError(Exception):
    """Raised when the configuration is incorrect"""


class Settings(Dynaconf):
    """Small extension of the settings object for our purposes

    Args:

        default_settings_filepath:
            Filepath to default settings file

        setting_files:
            A list containing setting files in ascending order according to
            importance (most important last).

        arguments:
            All positional arguments

        keyword_arguments:
            All keyword arguments

    """

    def __init__(
        self,
        default_settings_filepath,
        *arguments,
        settings_files: list[str] | None = None,
        **keyword_arguments,
    ) -> None:

        if settings_files is None:
            settings_files = []

        settings_files = [
            default_settings_filepath,
            ConfigurationUtility.site_config_filepath,
            ConfigurationUtility.user_config_filepath,
        ] + settings_files

        super().__init__(
            settings_files=settings_files,
            *arguments,
            **keyword_arguments,
        )
        self.validate_settings()

    def validate_settings(self) -> None:
        """Check settings for errors"""

        self.validators.register(
            *commands_validators(),
            *sensor_node_validators(),
            *sth_validators(),
            *stu_validators(),
        )

        try:
            self.validators.validate()
        except ValidationError as error:
            raise SettingsIncorrectError(
                f"Incorrect configuration: {error}"
            ) from error

    def acceleration_sensor(self):
        """Get the settings for the current acceleration sensor

        Returns:

            A configuration object for the currently selected accelerometer
            sensor

        """

        sensor_settings = self.sth.acceleration_sensor
        return sensor_settings[sensor_settings.sensor]


# -- Attributes ---------------------------------------------------------------


with as_file(
    files("icotest.config").joinpath("config.yaml")
) as repo_settings_filepath:
    try:
        settings = Settings(default_settings_filepath=repo_settings_filepath)
    except SettingsIncorrectError as settings_incorrect_error:
        handle_incorrect_settings(f"{settings_incorrect_error}")
    except (ParserError, ScannerError) as parsing_error:
        handle_incorrect_settings(
            f"Unable to parse configuration: {parsing_error}"
        )



"""Shared code for STU"""

# -- Imports ------------------------------------------------------------------

from logging import getLogger
from icotronic.can import STH

from icotest.config import settings

# -- Functions ----------------------------------------------------------------


async def read_self_test_voltages(sth: STH) -> tuple[float, float]:
    """Read acceleration voltages before, at and after self test

    Args:

        sth:

            The STH where the measurement should take place
    Returns:

        A tuple containing the **absolute** i.e. positive difference:

        - between the voltage before and at the self test and

        - the value of the voltage before and after the self test

        in Millivolt.

    """

    logger = getLogger(__file__)

    sensor = settings.acceleration_sensor()
    dimension = sensor.self_test.dimension
    reference_voltage = sensor.reference_voltage

    voltage_before_test = await sth.get_acceleration_voltage(
        dimension, reference_voltage
    )
    logger.info("Voltage before test: %s V", voltage_before_test)

    await sth.activate_acceleration_self_test(dimension)
    voltage_at_test = await sth.get_acceleration_voltage(
        dimension, reference_voltage
    )
    logger.info("Voltage at test: %s V", voltage_at_test)

    await sth.deactivate_acceleration_self_test(dimension)
    voltage_after_test = await sth.get_acceleration_voltage(
        dimension, reference_voltage
    )
    logger.info("Voltage after test: %s V", voltage_after_test)

    voltage_diff = voltage_at_test - voltage_before_test
    voltage_diff_abs = abs(voltage_diff)
    voltage_diff_before_after = abs(voltage_before_test - voltage_after_test)

    return (voltage_diff_abs * 1000, voltage_diff_before_after * 1000)
